// From the software distribution accompanying the textbook
// "A Practical Introduction to Data Structures and Algorithm Analysis,
// Third Edition" by Clifford A. Shaffer, Prentice Hall, 2007.
// Source code Copyright (C) 2006 by Clifford A. Shaffer.

// Huffman coding tree example program.
// First, read from the data file a set of strings and associated
// frequencies. These are placed onto a list of (single node) Huffman trees.
// Next, build a single Huffman coding tree for the set.  The strings and
// their codes are then output, with CodeTable storing the coding for
// each input string. Next, read commands to "encode" or "decode" strings,
// providing the appropriate output.

// Sample input files are supplied with this distribution, they are
// named "*.huff".

// This implementation uses a composite design
// This shows up mostly in the encode/decode functions, which now
// belong to the node methods.
// In contrast, see huffman which implements a "tree in control" design,
// and so needs explict tree-level encode/decode functions.

#include <iostream>
using namespace std;

#include "book.h"

// A simple class that stores a letter/frequency pair
// These are the data objects for the huffman tree
#include "freqpair.h"

// Overload for the FreqPair << operator
template <class Elem>
ostream& operator << (ostream& s, FreqPair<Elem>* pair)
  // Assume that the Freqpair element is printable
  { return(s << pair->val() << "|" << pair->weight()); }

// Include definition for sorted linked lists.
// We will use them to keep track of the forest of huffman trees during the
// building operation
#include "sllist.ht"

#define MAXCODELEN 20       // Max length of a huffman code
#define CODETABLELEN 100    // Maximum number of codes storable

// CodeTable maps objects to their associated codes.
template <class Elem, class Compare>
class CodeTable {
private:
  Elem* obs;             // Objects
  char** codes;  // Associated code values
  int currsize;  // Current number of objects in table
  int maxsize;  // Max objects permitted in table
public:
  CodeTable(int size) {
    obs = new Elem[size];
    codes = new char*[size];
	for (int i = 0; i<size; i++) {
	  codes[i] = new char[MAXCODELEN+1];
      for(int j=0; j<=MAXCODELEN; j++)
		  codes[i][j] = '-';
	  codes[i][MAXCODELEN] = '\0';
	}
	  maxsize = size;  currsize = 0;
  }
  void addobject(Elem& obj) {
    Assert(currsize < maxsize, "CodeTable is full!");
    obs[currsize++] = obj;
  }
  char* getcode(Elem& obj) {
	for (int i=0; i<currsize; i++)
	  if(Compare::eq(obj, obs[i])) return codes[i];
	return NULL;
  }
};  

#include "huffnode1.ht"

template <class Elem, class Comp>
ostream& operator << (ostream& s, HuffNode<Elem, Comp>* z)
{
  if (z->isLeaf())
    return s << ((LeafNode<Elem, Comp>*)z)->val();
  else
    return s << ((IntlNode<Elem, Comp>*)z)->weight();
}

#include "hufftree1.ht"

// Overload for the HuffTree << operator
template <class Elem, class Comp>
ostream& operator << (ostream& s, HuffTree<Elem, Comp>* z)
  { return s << z->weight(); }


// Comparator class for comparing two char variables
class ccCompare {
public:
  static bool lt(char x, char y) { return x < y; }
  static bool eq(char x, char y) { return x == y; }
  static bool gt(char x, char y) { return x > y; }
};


// Read the list of frequencies, make the forest, and set the
// list of entries into the code table.
void read_freqs(SLList<HuffTree<char, ccCompare>*,
				       HHCompare<char, ccCompare> >* forest,
                CodeTable<char, ccCompare>* ct, FILE* fp)
{ // Read a list of strings and frequencies from standard input,
  // building a list of Huffman coding tree nodes
  char buff[100];
  char buff2[100];
  char *ptr;
  char *ptr2;
  int freq;
  HuffTree<char, ccCompare>* temptree;

  while (true) {
    Assert(fgets(buff, 99, fp) != NULL,   // Read the next entry
           "No more codes to read -- where is end?");

    // process the entry, creating a new HuffTree
    for(ptr=buff; *ptr==' '; ptr++);  // Read first word
    if (strncmp(ptr, "end", 3) == 0)  // OK, we're done reading entries
      return;
    Assert(*ptr == '"', "First char was not a quote mark.");
    for (ptr2=buff2,ptr++; *ptr!='"'; ptr++)
      *ptr2++ = *ptr;
    *ptr2 = '\0'; // End of string
    for (ptr++; *ptr==' '; ptr++);
    Assert(isdigit(*ptr) != 0, "Must be a digit here.");
    freq = atoi(ptr);
    ct->addobject(buff2[0]);
    temptree = new HuffTree<char, ccCompare>(buff2[0], freq);

    // put in the list in sorted order
    // WARNING: This might be considered buggy. Nodes with equal weight will
    // be put in reverse order of appearance, not in alphabetical order.
    forest->insert(temptree);
  }
}

// Build Huffman tree from list fl
template <class Elem, class Comp>HuffTree<Elem, Comp>*
buildHuff(SLList<HuffTree<Elem, Comp>*,
          HHCompare<Elem, Comp> >* fl) {
  HuffTree<Elem, Comp> *temp1, *temp2, *temp3;
  for (fl->setStart(); fl->leftLength()+fl->rightLength()>1;
       fl->setStart()) { // While at least two items left
    fl->remove(temp1);   // Pull first two trees  
    fl->remove(temp2);   //   off the list
    temp3 = new HuffTree<Elem, Comp>(temp1, temp2);
    fl->insert(temp3);   // Put the new tree back on list
    delete temp1;        // Must delete the remnants
    delete temp2;        //   of the trees we created
  }
  return temp3;
}


void do_commands(HuffTree<char, ccCompare>* theTree,
                 CodeTable<char, ccCompare>* theTable, FILE *fp)
{
  int currchar;
  char buff[80];

  while (fgets(buff, 99, fp)) {
    if (strncmp(buff, "decode", 6) == 0) {
      for (currchar=0; buff[currchar] != '"'; currchar++);
      cout << "Decode " << &buff[currchar++];
	  while (buff[currchar] != '"') {
		int cnt = 0;
		char msg;
        theTree->decode(&buff[currchar], msg, cnt);
        cout << msg << endl;
		currchar += cnt;
	  }
    }
    else if(strncmp(buff, "encode", 6) == 0) {
      for (currchar=0; buff[currchar] != '"'; currchar++);
      cout << "Encode " << &buff[currchar++];
      for(; buff[currchar] != '"' ; currchar++) // Assume codes are characters.  Should generalize this.
        if (buff[currchar] == ' ') cout << ' ';
        else cout << theTable->getcode(buff[currchar]);
    }
    cout << "\n";
  }
}


// Driver for testing Huffman tree code
main(int argc, char** argv)
{
  // This list holds the Huffman forest during construction
  SLList<HuffTree<char, ccCompare>*, HHCompare<char, ccCompare> >* forest =
      new SLList<HuffTree<char, ccCompare>*, HHCompare<char, ccCompare> >;
  // This will be the eventual Huffman tree
  HuffTree<char, ccCompare>* theTree;
  CodeTable<char, ccCompare>* theTable =
	   new CodeTable<char, ccCompare>(CODETABLELEN);
  // Working storage for the tree traversal that builds the code table
  char prefix[MAXCODELEN+1];
  // total is used to calculate the average code length
  double total = 0;
  FILE *fp;  // The file pointer

  // Check command line parameter for frequency file
  Assert(argc == 2, "Usage: huffman <frequency file>");
  Assert((fp = fopen(argv[1], "rt")) != NULL, "No such file");

  // Now, read in the list of frequencies, and initialize the
  //   forest of Huffman trees.
  cout << "Read frequencies\n";
  read_freqs(forest, theTable, fp);

  forest->print();

  // Now, build the tree.
  cout << "Build the tree\n";
  theTree = buildHuff(forest);

  // Now, output the tree, which also creates the code table.
  cout << "Output the tree\n";
  theTree->buildcode(theTable, prefix, 0, total);
  cout << "Average code length is "
       << total/(double)theTree->weight() << "\n";

  // Finally, do the encode/decode commands to test the system.
  do_commands(theTree, theTable, fp);
  return 0;
}
